# Copyright © 2021-2022 Arrikto Inc.  All Rights Reserved.

"""Kiwi example using the Kale SDK.

This script creates an ML pipeline that classifies dog images
and runs in parallel on the same Kiwi-enabled GPU
"""

from kale.types import MarshalData
import tensorflow as tf

from tensorflow.keras.preprocessing.image import ImageDataGenerator
from rok_kubernetes.models import (V1Affinity, V1PodAffinity, V1LabelSelector,
                                   V1PodAffinityTerm)

from kale.sdk import pipeline, step
from PIL import ImageFile

ImageFile.LOAD_TRUNCATED_IMAGES = True

affinity = V1Affinity(pod_affinity=V1PodAffinity(
                      required_during_scheduling_ignored_during_execution=[
                          V1PodAffinityTerm(
                              topology_key="kubernetes.io/hostname",
                              label_selector=V1LabelSelector(
                                  match_labels={"app": "dogbreed"}))]))


@step(name="load-data")
def load_data(img_size: int = 224, batch_size: int = 32):
    def get_train_generator():
        data_datagen = ImageDataGenerator(rescale=1. / 255,
                                          width_shift_range=.2,
                                          height_shift_range=.2,
                                          brightness_range=[0.5, 1.5],
                                          horizontal_flip=True)
        return data_datagen.flow_from_directory(
            "dogImages/train/", target_size=(img_size, img_size),
            batch_size=batch_size)

    def get_valid_generator():
        data_datagen = ImageDataGenerator(rescale=1. / 255)
        return data_datagen.flow_from_directory(
            "dogImages/valid/", target_size=(img_size, img_size),
            batch_size=batch_size)

    def get_test_generator():
        data_datagen = ImageDataGenerator(rescale=1. / 255)
        return data_datagen.flow_from_directory(
            "dogImages/test/", target_size=(img_size, img_size),
            batch_size=batch_size)

    return get_train_generator, get_valid_generator, get_test_generator


@step(name="custom-classifier",
      deploy_config={"limits": {"arrikto.com/gpu": "1"}})
def custom_classifier(get_train_generator: MarshalData,
                      get_valid_generator: MarshalData, epochs: int = 2,
                      img_size: int = 224, number_of_nodes: int = 512,
                      lr: float = 0.001):
    model = tf.keras.models.Sequential(
        [tf.keras.layers.Conv2D(16, 3, activation="relu",
                                input_shape=(img_size, img_size, 3)),
         tf.keras.layers.MaxPool2D(),
         tf.keras.layers.Conv2D(32, 3, activation="relu"),
         tf.keras.layers.MaxPool2D(),
         tf.keras.layers.Conv2D(64, 3, activation="relu"),
         tf.keras.layers.MaxPool2D(),
         tf.keras.layers.GlobalAveragePooling2D(),
         tf.keras.layers.Dense(int(number_of_nodes), activation="relu"),
         tf.keras.layers.Dense(133, activation="softmax")])

    model.compile(
        optimizer=tf.optimizers.Adam(learning_rate=float(lr)),
        loss=tf.losses.categorical_crossentropy,
        metrics=["accuracy"]
    )

    train_generator = get_train_generator()
    valid_generator = get_valid_generator()

    tb_callback = tf.keras.callbacks.TensorBoard(
        log_dir="custom_classifier_logs")

    model.fit(train_generator, epochs=epochs, validation_data=valid_generator,
              callbacks=[tb_callback])


@step(name="vgg16-classifier",
      deploy_config={"limits": {"arrikto.com/gpu": "1"}})
def vgg16_classifier(get_train_generator: MarshalData,
                     get_valid_generator: MarshalData, epochs: int = 2,
                     img_size: int = 224, lr: float = 0.001):
    vgg_body = tf.keras.applications.VGG16(
        weights="imagenet",
        include_top=False,
        input_shape=(img_size, img_size, 3)
    )

    vgg_body.trainable = False

    inputs = tf.keras.layers.Input(shape=(img_size, img_size, 3))
    x = vgg_body(inputs, training=False)
    x = tf.keras.layers.GlobalAveragePooling2D()(x)
    outputs = tf.keras.layers.Dense(133, activation="softmax")(x)
    vgg_model = tf.keras.Model(inputs, outputs)

    vgg_model.summary()

    vgg_model.compile(
        optimizer=tf.optimizers.Adam(learning_rate=float(lr)),
        loss=tf.losses.categorical_crossentropy,
        metrics=["accuracy"]
    )

    train_generator = get_train_generator()
    valid_generator = get_valid_generator()

    vgg_model.fit(train_generator, epochs=epochs,
                  validation_data=valid_generator)


@step(name="resnet50-classifier",
      deploy_config={"limits": {"arrikto.com/gpu": "1"}})
def resnet50_classifier(get_train_generator: MarshalData,
                        get_valid_generator: MarshalData, epochs: int = 2,
                        img_size: int = 224, lr: float = 0.001):
    resnet_body = tf.keras.applications.ResNet50V2(
        weights="imagenet",
        include_top=False,
        input_shape=(img_size, img_size, 3)
    )

    resnet_body.trainable = False

    inputs = tf.keras.layers.Input(shape=(img_size, img_size, 3))

    x = resnet_body(inputs, training=False)
    x = tf.keras.layers.Flatten()(x)
    outputs = tf.keras.layers.Dense(133, activation="softmax")(x)

    resnet_model = tf.keras.Model(inputs, outputs)

    resnet_model.compile(
        optimizer=tf.optimizers.Adam(learning_rate=float(lr)),
        loss=tf.losses.categorical_crossentropy,
        metrics=["accuracy"]
    )

    train_generator = get_train_generator()
    valid_generator = get_valid_generator()

    resnet_model.fit(train_generator, epochs=epochs,
                     validation_data=valid_generator)


@pipeline(name="dog-classifier", experiment="kiwi-tutorial",
          deploy_config={"affinity": affinity, "labels": {"app": "dogbreed"}},
          autosnapshot=False)
def dog_classifier(epochs: int = 2, lr: float = 6e-4, batch_size: int = 32,
                   img_size: int = 224, number_of_nodes: int = 512):
    train_generator, valid_generator, junk = load_data(img_size, batch_size)
    custom_classifier(train_generator, valid_generator, epochs, img_size,
                      number_of_nodes, lr)
    vgg16_classifier(train_generator, valid_generator, epochs, img_size, lr)
    resnet50_classifier(train_generator, valid_generator, epochs, img_size, lr)


if __name__ == "__main__":
    dog_classifier(epochs=4, lr=6e-4, batch_size=32,
                   img_size=224, number_of_nodes=512)
